import fiona
from shapely import wkt, geometry
from wq.io.loaders import FileLoader
from wq.io.parsers.base import BaseParser
from wq.io.mappers import TupleMapper


class FionaLoaderParser(FileLoader, BaseParser):
    """
    Composite loader & parser mixin for GIS data, powered by Fiona
    """
    layer_id = None
    meta = {}
    key_field = 'id'

    def load(self):
        # Load and parse the dataset at the same time via Fiona
        pass

    def parse(self):
        try:
            layers = fiona.listlayers(self.filename)
        except (ValueError, IOError) as e:
            self.data = []
            return

        # If multiple layers, parse all of them (!)
        if len(layers) > 1 and self.layer_id is None:
            cls = type(self)
            self.data = [{
                'id': id,
                'name': name,
                'data': cls(filename=self.filename, layer_id=id)
            } for id, name in enumerate(layers)]
        else:
            # One layer, load & parse GIS data
            with fiona.open(self.filename, layer=self.layer_id) as f:
                self.meta = f.meta
                self.data = map(self.parse_feature, f)

    def parse_feature(self, f):
        # Flatten Fiona's GeoJSON-style representation into something more
        # amenable to namedtuple-ing
        feat = {key: value for key, value in f['properties'].items()}
        if 'id' not in feat and 'ID' not in feat:
            feat['id'] = f['id']
        feat['geometry'] = f['geometry']
        return feat

    def dump_feature(self, feat, i):
        # Undo aforementioned flattening
        return {
            'id': feat.get('id', feat.get('ID', i)),
            'geometry': feat['geometry'],
            'properties': {
                key: value for key, value in feat.items()
                if key not in ('geometry', 'id',)
            }
        }

    def dump(self):
        # Dump and save the dataset at the same time via Fiona
        pass

    def save(self):
        with fiona.open(self.filename, 'w', **self.meta) as f:
            for i, feat in enumerate(self.data):
                f.write(self.dump_feature(feat, i))


class ShapeMapper(TupleMapper):
    """
    Map Fiona's GeoJSON-style geometries to and from Shapely shapes
    """
    def map_value(self, field, value):
        value = super(ShapeMapper, self).map_value(field, value)
        if field == 'geometry':
            value = geometry.shape(value)
        return value

    def unmap_value(self, field, value):
        if field == 'geometry':
            value = geometry.mapping(value)
        return super(ShapeMapper, self).unmap_value(field, value)


class WktMapper(ShapeMapper):
    """
    Map geometries to and from WKT (good for Django integration)
    """
    def map_value(self, field, value):
        value = super(WktMapper, self).map_value(field, value)
        if field == 'geometry':
            value = wkt.dumps(value)
        return value

    def unmap_value(self, field, value):
        if field == 'geometry':
            value = wkt.loads(value)
        return super(WktMapper, self).unmap_value(field, value)



from wq.io.mappers import TupleMapper
from .mixins import FionaLoaderParser, ShapeMapper, WktMapper
from wq.io.base import BaseIO


class MetaSyncIO(BaseIO):
    """
    Custom sync() to handle transfering Fiona meta
    """
    def sync(self, other, save=True):
        other.meta = self.meta
        super(MetaSyncIO, self).sync(other, save)

    def get_field_names(self):
        if self.field_names is None and self.meta is not None:
            return (
                ['id', 'geometry']
                + self.meta['schema']['properties'].keys()
            )
        return super(MetaSyncIO, self).get_field_names()


class GisIO(FionaLoaderParser, TupleMapper, MetaSyncIO):
    pass


class ShapeIO(FionaLoaderParser, ShapeMapper, MetaSyncIO):
    pass


class WktIO(FionaLoaderParser, WktMapper, MetaSyncIO):
    pass



from collections import MutableMapping, MutableSequence


class BaseIO(MutableMapping, MutableSequence):
    "wq.io.BaseIO: Base class for generic resource management"

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self.refresh()

    def refresh(self):
        self.load()
        self.parse()

    def load(self):
        "Open a resource (defined by loader mixins)"
        # self.file = ...
        pass

    def parse(self):
        """
        Parse a resource (defined by parser mixins).
        Result should be an iterable of dicts.
        """
        # self.data = some_parse_method(self.file)
        pass

    def dump(self, file=None):
        ""
        if file is None:
            file = self.file
        file.write(unicode(self.data))

    def save(self):
        ""
        self.dump(self.file)

    field_names = None

    def get_field_names(self):
        "Returns a list of raw fields to expect (defined by parser mixins)"
        if self.field_names is not None:
            #Support specifying field_names as string (like namedtuple does)
            if isinstance(self.field_names, basestring):
                return self.field_names.replace(',', ' ').split()
            else:
                return self.field_names

        # If no defined field names, try to retrieve from first record
        if getattr(self, 'data', None) and len(self.data) > 0:
            return self.data[0].keys()
        else:
            return None

    @property
    def key_field(self):
        "Assign a key_field to use the resource as a Map"
        return None

    def get_key_field(self):
        return self.key_field

    def usable_item(self, item):
        "Hook to allow items to be transformed"
        return item

    def parse_usable_item(self, uitem):
        "Hook to allow items to be untransformed"
        return uitem

    def find_index(self, key):
        pk = self.get_key_field()
        if pk is None:
            return key
        for index, item in enumerate(self.data):
            uitem = self.usable_item(item)
            if isinstance(uitem, dict):
                if uitem.get(pk, None) == key:
                    return index
            else:
                if getattr(uitem, pk, None) == key:
                    return index

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        index = self.find_index(key)
        if index is None:
            raise KeyError
        return self.usable_item(self.data[index])

    def __setitem__(self, key, uitem):
        item = self.parse_usable_item(uitem)
        index = self.find_index(key)
        if index is not None:
            self.data[index] = item
        else:
            self.data.append(item)

    def __delitem__(self, key):
        index = self.find_index(key)
        if index is None:
            raise KeyError
        del self.data[index]

    def insert(self, index, uitem):
        item = self.parse_usable_item(uitem)
        self.data.insert(index, item)

    def __iter__(self):
        for item in self.data:
            uitem = self.usable_item(item)
            if uitem is None:
                return
            pk = self.get_key_field()
            if pk is None:
                yield uitem
            elif isinstance(uitem, dict):
                yield uitem.get(pk, None)
            else:
                yield getattr(uitem, pk, None)

    def sync(self, other, save=True):
        if self.get_key_field() is None or other.get_key_field() is None:
            raise Exception("Key field required to sync!")
        for key in self:
            other[key] = self[key]
        if save:
            other.save()

    def copy(self, other, save=True):
        del other.data[:]
        for item in self.data:
            uitem = self.usable_item(item)
            other.append(uitem)
        if save:
            other.save()



import os
from setuptools import setup, find_packages


LONG_DESCRIPTION = """
Consistent iterable API for reading and writing to external datasets
"""


def parse_markdown_readme():
    """
    Convert README.md to RST via pandoc, and load into memory
    (fallback to LONG_DESCRIPTION on failure)
    """
    # Attempt to run pandoc on markdown file
    import subprocess
    try:
        subprocess.call(
            ['pandoc', '-t', 'rst', '-o', 'README.rst', 'README.md']
        )
    except OSError:
        return LONG_DESCRIPTION

    # Attempt to load output
    try:
        readme = open(os.path.join(
            os.path.dirname(__file__),
            'README.rst'
        ))
    except IOError:
        return LONG_DESCRIPTION
    return readme.read()


def get_version():
    version = open("version.py").read().strip()
    version = version.replace("VERSION = ", '')
    version = version.replace('"', '')
    return version


def create_wq_namespace():
    """
    Generate the wq namespace package
    (not checked in, as it technically is the parent of this folder)
    """
    if os.path.isdir("wq"):
        return
    os.makedirs("wq")
    init = open(os.path.join("wq", "__init__.py"), 'w')
    init.write("__import__('pkg_resources').declare_namespace(__name__)")
    init.close()


def find_wq_packages(submodule):
    """
    Add submodule prefix to found packages.  The packages within each wq
    submodule exist at the top level of their respective repositories.
    """
    packages = ['wq', submodule]
    package_dir = {submodule: '.'}
    for package in find_packages():
        if package == 'wq':
            continue
        full_name = submodule + "." + package
        packages.append(full_name)
        package_dir[full_name] = package.replace('.', os.sep)

    return packages, package_dir


create_wq_namespace()
packages, package_dir = find_wq_packages('wq.io')


setup(
    name='wq.io',
    version=get_version(),
    author='S. Andrew Sheppard',
    author_email='andrew@wq.io',
    url='http://wq.io/wq.io',
    license='MIT',
    packages=packages,
    package_dir=package_dir,
    namespace_packages=['wq'],
    description=LONG_DESCRIPTION.strip(),
    long_description=parse_markdown_readme(),
    install_requires=['httplib2', 'xlrd'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Scientific/Engineering :: GIS',
    ]
)
